"use strict";
/**
 * A thing that creates {@linkcode typedoc#DeclarationReflection} instances from parsed
 * command & execute method data.
 * @module
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.createReflections = void 0;
const typedoc_1 = require("typedoc");
const model_1 = require("../model");
/**
 * Creates and adds a child {@linkcode CommandReflection} to this reflection
 *
 * During "normal" usage of TypeDoc, one would call
 * `createDeclarationReflection()`. But since we've subclassed
 * `DeclarationReflection`, we cannot call it directly.  It doesn't seem to do
 * anything useful besides instantiation then delegating to
 * `postReflectionCreation()`; so we just need to call it directly.
 *
 * Finally, we call `finalizeDeclarationReflection()` which I think just fires
 * some events for other plugins to potentially use.
 * @param log Logger
 * @param data Command reference
 * @param route Route
 * @param parent Commands reflection
 * @internal
 */
function createCommandReflection(log, ctx, data, parent, route) {
    const commandRefl = new model_1.CommandReflection(data, parent, route);
    // yes, the `undefined`s are needed
    ctx.postReflectionCreation(commandRefl, undefined, undefined);
    ctx.finalizeDeclarationReflection(commandRefl);
}
/**
 * Create a new {@linkcode CommandsReflection} and all {@linkcode CommandReflection} children within it.
 * @param log - Logger
 * @param ctx - Current context
 * @param parent - Parent module (or project)
 * @param commandInfo - Command information for `module`
 * @internal
 */
function createCommandsReflection(log, ctx, parent, commandInfo) {
    // TODO: parent.name may not be right here
    const commandsRefl = new model_1.CommandsReflection(parent.name, ctx.project, commandInfo);
    /**
     * See note in `#createCommandReflection` above about this call
     */
    ctx.postReflectionCreation(commandsRefl, undefined, undefined);
    const parentCtx = ctx.withScope(commandsRefl);
    const { routeMap: routeMap, execMethodDataSet: execCommandsData } = commandInfo;
    // sort routes in alphabetical order
    const sortedRouteMap = new Map([...routeMap.entries()].sort());
    for (const [route, commandMap] of sortedRouteMap) {
        for (const data of commandMap.values()) {
            createCommandReflection(log, parentCtx, data, commandsRefl, route);
        }
    }
    // sort execute commands in alphabetical order
    const sortedExecCommandsData = new Set([...execCommandsData].sort());
    for (const data of sortedExecCommandsData) {
        createCommandReflection(log, parentCtx, data, commandsRefl);
    }
    ctx.finalizeDeclarationReflection(commandsRefl);
    return commandsRefl;
}
/**
 * Creates custom {@linkcode typedoc#DeclarationReflection}s from parsed command & execute method data.
 *
 * These instances are added to the {@linkcode Context} object itself; this mutates TypeDoc's internal state. Nothing is returned.
 * @param ctx TypeDoc Context
 * @param parentLog Plugin logger
 * @param commandInfo Command info from converter; a map of parent reflections to parsed data
 */
function createReflections(ctx, parentLog, commandInfo) {
    const log = parentLog.createChildLogger('builder');
    const { project } = ctx;
    // note that this could be an empty array
    const modules = project.getChildrenByKind(typedoc_1.ReflectionKind.Module);
    // the project itself may have commands, as well as any modules within the project
    let hasCommands = false;
    for (const parent of [...modules, project]) {
        if (commandInfo.get(parent)?.hasData) {
            hasCommands = true;
            createCommandsReflection(log, ctx, parent, commandInfo.get(parent));
        }
    }
    if (!hasCommands) {
        log.warn('No Appium commands found in the entire project');
        // TODO: maybe we should abort processing gracefully here? or throw?
    }
}
exports.createReflections = createReflections;
//# sourceMappingURL=builder.js.map